/*
  vpd_util - VPD support routines.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: util.c,v 1.26 2005/02/18 05:04:05 martins Exp $
*/
  
/* For pread, pwrite: */
#define _XOPEN_SOURCE 500

/* For asprintf: */
#define _GNU_SOURCE

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

#include "util.h"

char * vpd_program_name = "ERROR";

void
die(const char * message)
{
	fprintf(stderr, "%s: %s\n", vpd_program_name, message);
	exit(1);
}

void
die2(const char * message, const char * sub_message)
{
	fprintf(stderr, "%s: %s (%s)\n",
		vpd_program_name, message, sub_message);
	exit(1);
}

char *
build_relative_filename(const char * dir,
			const char * relpath)
{
	char * ret;

	if (-1 == asprintf(&ret, "%s/%s", dir, relpath)) {
		ret = NULL;
	}
	return ret;
}

bool
read_to_eof(FILE   * file,
	    char  ** buf,
	    size_t * length,
	    int    * err)
{
	size_t n;
	size_t done = 0;

	const size_t chunk_size = 4096;
	int chunk_count = 1;

	if (NULL == (*buf = malloc(chunk_size * chunk_count)))
		goto bail;
	
	/* - 1 everywhere because we're going to bonus NUL-terminate. */
	while (0 < (n = fread(*buf + done, 1,
			      (chunk_size * chunk_count - 1) - done, file))) {
		done += n;
		if (done == (chunk_size * chunk_count - 1)) {
			chunk_count++;
			if (NULL ==
			    (*buf = realloc(*buf, chunk_size * chunk_count)))
				goto bail;
		}
	}
	
	if (0 != ferror(file))
		goto bail;

	(*buf)[done] = '\0';

	if (NULL != length) {
		*length = done;
	}

	return true;

 bail:
	if (NULL != length)
		*length = 0;
	if (NULL != err)
		*err = errno;
	if (NULL != *buf)
		free(*buf);
	*buf = NULL;
	return false;
}

bool
file_read_contents(const  char   * pathname,
		          char  ** buf,
		          size_t * length,
		          int    * err)
{
	FILE * file;
	char * cmd = NULL;
	bool ret;

	*buf = NULL;

	if (NULL == pathname)
		file = stdin;
	else {
		if (strrchr(pathname, '|') !=
		    (pathname + strlen(pathname) - 1)) {
			if (NULL == (file = fopen(pathname, "r")))
				goto bail;
		} else {
			/* I don't think this is used anywhere. */
			cmd = strdup(pathname);
			cmd[strlen(pathname) - 1] = '\0';
			if (NULL == (file = popen(cmd, "r")))
				goto bail;
		}
	}

	ret = read_to_eof(file, buf, length, err);

	if (NULL != cmd) {
		free(cmd);
		pclose(file);
	} else
		fclose(file);

	return ret;

 bail:
	if (NULL != length)
		*length = 0;
	if (NULL != err)
		*err = errno;
	if (NULL != cmd)
		free(cmd);
	return false;
}

bool
file_relative_read_contents(const char   * dir,
			    const char   * relpath,
			    char  ** buf,
			    size_t * length,
			    int    * err)
{
	bool ret;
	char * file;

	ret = false;

	if (NULL != (file = build_relative_filename(dir, relpath))) {
		if (file_read_contents(file, buf, length, err))
			ret = true;
		free(file);
	}

	return ret;
}

bool
file_write_contents(const  char  * pathname,
		    const  char  * buf,
		           size_t  count,
		           int   * err)
{
	bool ret;
	int fd;
	size_t n;
	size_t done = 0;

	ret = false;

	if (-1 != (fd = open(pathname, O_CREAT|O_WRONLY|O_TRUNC, 0644))) {
		ret = true;
		while (done != count) {
			if (-1 == (n = write(fd, buf + done, count - done))) {
				*err = errno;
				ret = false;
				break;
			}
				
			done += n;
		}
		
		close(fd);
	} else
		*err = errno;

	return ret;
}

bool
file_relative_write_contents(const char * dir,
			     const char * relpath,
			     const char * buf,
			     size_t   count,
			     int * err)
{
	bool ret;
	char * file;

	ret = false;

	if (NULL != (file = build_relative_filename(dir, relpath))) {
		if (file_write_contents(file, buf, count, err))
			ret = true;
		free(file);
	}

	return ret;
}

/* These should never give short reads or writes */
ssize_t
big_cool_pread(int fd, uint8_t *buf, size_t count, off_t offset)
{
	ssize_t total = 0;
	ssize_t last  = 0;

	do {
		last = pread(fd, buf + total, count - total, offset + total);
		if (-1 != last) {
			total += last;
		}
	} while ((0 < last) && (total < count));
	
	if ((-1 == last) && (0 == total)) {
		total = -1;
	}

	return total;
}

ssize_t
big_cool_pwrite(int fd, const uint8_t *buf, size_t count, off_t offset)
{
	ssize_t total = 0;
	ssize_t last  = 0;

	do {
		last = pwrite(fd, buf + total, count - total, offset + total);
		if (-1 != last) {
			total += last;
		}
	} while ((0 < last) && (total < count));
	
	if ((-1 == last) && (0 == total)) {
		total = -1;
	}

	return total;
}

const char *
string_array_lookup(const struct string_array * a,
		    const char * d,
		    int k)
{
	int i = 0;
	while (a[i].val != NULL) {
		if (a[i].key == k) {
			return a[i].val;
		}
		i++;
	}
	return d;
}

char *
strndup_trim(const char * s,
	     size_t n)
{
	char * ret;
	int b;
	int i;
	int e;

	ret = NULL;

	/* Find beginning (b): first non-whitespace character. */
	b = 0;
	while ((b < n) && isspace(s[b])) {
		b++;
	}

	/* Find printables, remembering last non-space character at e. */
	e = i = b;
	while ((i < n) && isprint(s[i])) {
		if (!isspace(s[i]))
			e = i;
		i++;
	}

	/* Make e point past last useful character. */
	e++;

	/* Duplicate string between b and e. */
	if ((e > b) && (NULL != (ret = malloc(e - b + 1)))) {
		memcpy(ret, &s[b], e - b);
		ret[e - b] = '\0';
	}

	return ret;
}

bool
line_scanner_harness(const char * filename,
		     bool (* line_scanner) (const char *, void *),
		     void * result)
{
	bool ret;
	char * data;
	size_t len;
	int    err;
	char * this_line;
	char * next_line;
	bool more;

	ret = false;

	data = this_line = next_line = NULL;

	if (file_read_contents(filename, &data, &len, &err)) {
		more = true;
		this_line = data;
		while (more && (NULL != this_line)) {
			next_line = memchr(this_line, '\n',
					   len - (this_line - data));
			if (NULL != next_line) {
				*next_line = '\0';
				next_line++;
			}

			more = (*line_scanner)(this_line, result);

			if ((NULL != next_line) && ((next_line - data) < len))
				this_line = next_line;
			else
				this_line= NULL;
		}
		free(data);

		ret = true;
		
	}

	return ret;
}
