/*
  vpd - Utilities for in memory and disk-based VPD handling.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004, 2005
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: vpd.c,v 1.4 2005/02/22 22:28:28 martins Exp $
*/
  
#define _GNU_SOURCE

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "growbuf.h"
#include "util.h"
#include "vpd.h"


#define VPD_SUBDIR "linux,vpd"
#define VPD_LIST_FILE ".lsvpd"

vpd_list *
vpd_list_new(const vpd_key k, const char * v, size_t vlen)
{
	vpd_list * vl;

	vl = malloc(sizeof(vpd_list));
	if (NULL != vl) {
		vl->val = malloc(vlen + 1);
		if (NULL != vl->val) {
			memcpy(vl->val, v, vlen);
			vl->val[vlen] = '\0';
			memcpy(vl->key, k, sizeof(vl->key));
			INIT_LIST_HEAD(&vl->head);
		} else {
			free(vl);
			vl = NULL;
		}
	}

	return vl;
}

vpd_list *
vpd_list_new_empty()
{
	/* Redundant list head, ignored by list_for_each. */
	return vpd_list_new("**", "", 0);
}

void
vpd_list_free(vpd_list * vl)
{
	vpd_list * t1;
	vpd_list * t2;

	if (NULL != vl) {
		list_for_each_entry_safe(t1, t2, &vl->head, head) {
			free(t1->val);
			free(t1);
		}
		free(vl->val);
		free(vl);
	}
}


vpd_list_list *
vpd_list_list_new(vpd_list * vl)
{
	vpd_list_list * vll;

	vll = malloc(sizeof(*vll));
	if (NULL != vll) {
		vll->vpd_list = vl;
		INIT_LIST_HEAD(&vll->head);
	}

	return vll;
}

vpd_list_list *
vpd_list_list_new_empty()
{
	/* Redundant list head, ignored by list_for_each. */
	return vpd_list_list_new(NULL);
}

void
vpd_list_list_free(vpd_list_list * vll)
{
	vpd_list_list * t1;
	vpd_list_list * t2;

	if (NULL != vll) {
		list_for_each_entry_safe(t1, t2, &vll->head, head) {
			vpd_list_free(t1->vpd_list);
			free(t1);
		}
		free(vll);
	}
}

vpd_list *
vpd_list_append(vpd_list * vl, vpd_list * new)
{
	vpd_list * ret;

	ret = new;

	if (NULL != vl) {
		list_add_tail(&new->head, &vl->head);
		ret = vl;
	}

	return ret;
}

vpd_list *
vpd_list_prepend(vpd_list * vl, vpd_list * new)
{
	vpd_list * ret;

	ret = new;

	if (NULL != vl) {
		list_add(&new->head, &vl->head);
		ret = vl;
	}

	return ret;
}

vpd_list *
vpd_list_append_key_value(vpd_list * vl,
			  const vpd_key k, const char * v, size_t vlen)
{
	vpd_list * t;

	t = vpd_list_new(k, v, vlen);
	if (NULL != t)
		t = vpd_list_append(vl, t);

	return t;
}

vpd_list *
vpd_list_prepend_key_value(vpd_list * vl,
			  const vpd_key k, const char * v, size_t vlen)
{
	vpd_list * t;

	t = vpd_list_new(k, v, vlen);
	if (NULL != t)
		t = vpd_list_prepend(vl, t);

	return t;
}

vpd_list_list *
vpd_list_list_append(vpd_list_list * vll, vpd_list * vl)
{
	vpd_list_list * ret;

	ret = vpd_list_list_new(vl);
	if (NULL != ret)
		if (NULL != vll) {
			list_add_tail(&ret->head, &vll->head);
			ret = vll;
		}

	return ret;
}

bool
vpd_list_to_text(const vpd_list * vl, growbuf ** buf)
{
	bool ret;
	vpd_list * t;

	ret = false;

	if ((NULL != *buf) && (NULL != vl)) {
		list_for_each_entry(t, &vl->head, head) {
			*buf = growbuf_append(*buf, "*", 1);
			*buf = growbuf_append(*buf, t->key, sizeof(t->key));
			*buf = growbuf_append(*buf, " ", 1);
			*buf = growbuf_append(*buf, t->val, strlen(t->val));
			*buf = growbuf_append(*buf, "\n", 1);
		}
		ret = true;
	}

	return ret;
}

growbuf *
vpd_list_list_to_text(const vpd_list_list * vll)
{
	vpd_list_list * t;

	growbuf * ret = growbuf_new(512);

	if (NULL != vll)
		list_for_each_entry(t, &vll->head, head)
			if (!vpd_list_to_text(t->vpd_list, &ret)) {
				growbuf_free(ret);
				return NULL;
			}

	return ret;
}

static char *
vpd_key_value_to_vpd_dir(const char * vpd_dir, const vpd_key k, const char * v)
{
	char * kk;
	char * f;
	short i;
	int err;

	/*
	 * Could use an array for kk and snprintf, but then would have
	 * to strdup it at the end and that might fail after we've
	 * written the value to the file.
	 */
	f = NULL;
	kk = NULL;
	for (i = 0; i < 1000 ; i++) {
		if (0 == i) {
			if (-1 == asprintf(&kk, "%2.2s", k))
				kk = NULL;
		} else {
			if (-1 == asprintf(&kk, "%2.2s.%03d", k, i))
				kk = NULL;
		}

		if (NULL != kk) {
			f = build_relative_filename(vpd_dir, kk);
			if (NULL != f) {
				if (0 != access(f, F_OK))
					break;

				free(f);
				f = NULL;
			}
			
			free(kk);
			kk = NULL;
		}
	}

	if (NULL != f) {
		if (!file_write_contents(f, v, strlen(v), &err)) {
			free(kk);
			kk = NULL;
		}
		free(f);
	}

	return kk;
}

static void
vpd_dir_append_string_to_keys_file(FILE * ksfd, const char * s)
{
	fprintf(ksfd, "%s ", s);
}

bool
vpd_list_to_vpd_dir(const vpd_list * vl, const char * vpd_dir)
{
	vpd_list * t;
	bool ret = true;
	
	char * s;
	FILE * ksfd;
	char * kk;

	if (0 != mkdir(vpd_dir, 0777))
		return false;
		
	if (NULL == (s = build_relative_filename(vpd_dir, VPD_LIST_FILE)))
		return false;
	ksfd = fopen(s, "w");
	free(s);
	if (NULL == ksfd)
		return false;

	if (NULL != vl)
		list_for_each_entry(t, &vl->head, head) {
			kk = vpd_key_value_to_vpd_dir(vpd_dir, t->key, t->val);
			if (NULL != kk) {
				vpd_dir_append_string_to_keys_file(ksfd, kk);
				free(kk);
			} else
				ret = false;
		}
	fclose(ksfd);

	return ret;
}

bool
vpd_list_list_to_node(const vpd_list_list * vll, const char * node)
{
	vpd_list_list * t;

	short subnode = 0;
	char * l;
	char * d;

	bool ret = true;
	
	if (NULL == (l = build_relative_filename(node, VPD_SUBDIR)))
		return false;

	if (0 != mkdir(l, 0777)) {
		free(l);
		return false;
	}
		
	if (NULL != vll)
		list_for_each_entry(t, &vll->head, head)
			if (-1 != asprintf(&d, "%s/%03d", l, subnode)) {
				ret = ret && \
					vpd_list_to_vpd_dir(t->vpd_list, d);
				free(d);
				subnode++;
			} else
				ret = false;

	free(l);

	return ret;
}

vpd_list *
vpd_text_to_vpd_list(const growbuf * buf)
{
	unsigned char * t1;
	unsigned char * t2;
	unsigned char * t3;

	vpd_list * vl;
	growbuf_iter * i;

	vl = vpd_list_new_empty();

	i = growbuf_iter_new(buf);
	if (NULL == i) {
		vpd_list_free(vl);
		return NULL;
	}

	while (growbuf_iter_have(i, 4)) {
		growbuf_iter_next(i);
		t1 = growbuf_iter_current(i);
		growbuf_iter_move(i, 2);
		t2 = growbuf_iter_current(i);
		growbuf_iter_next(i);
		if (!growbuf_iter_more(i))
			break;

		t2 = growbuf_iter_current(i);
		if (NULL == (t3 = memchr(t2, '\n', growbuf_iter_remaining(i))))
			break;

		vl = vpd_list_append_key_value(vl, t1, t2, (size_t) (t3 - t2));
		growbuf_iter_move(i, (off_t) (((unsigned char *) t3) - t2) +1);
	}

	growbuf_iter_free(i);

	return vl;
}

static vpd_list_list *
vpd_list_to_vpd_list_list(const vpd_list * vl, const vpd_key sep)
{
	vpd_list * t;

	vpd_list * c = NULL;

	vpd_list_list * ret = vpd_list_list_new_empty();

	if (NULL != vl)
		list_for_each_entry(t, &vl->head, head) {
			if (0 == memcmp(sep, t->key, sizeof(vpd_key))) {
				if (NULL != c)
					ret = vpd_list_list_append(ret, c);
				c = vpd_list_new_empty();
			}
			
			if (NULL == c) {
				vpd_list_list_free(ret);
				break;	
			}
			
			c = vpd_list_append_key_value(c, t->key,
						      t->val, strlen(t->val));
		}

	if ((NULL != c) && (NULL != ret))
		ret = vpd_list_list_append(ret, c);

	return ret;
}

vpd_list_list *
vpd_text_to_vpd_list_list(const growbuf * buf, const vpd_key sep)
{
	/* There are marginally more efficient ways of doing this, but
	 * this is by far the simplest... */

	vpd_list * t = vpd_text_to_vpd_list(buf);

	vpd_list_list * ret = vpd_list_to_vpd_list_list(t, sep);

	vpd_list_free(t);

	return ret;
}

char *
vpd_dir_get_subdirnum(const char * dir, int subdirnum)
{
	char * ret;

	ret = NULL;
	if (-1 == asprintf(&ret, "%s/%s/%03d", dir, VPD_SUBDIR, subdirnum))
		ret = NULL;
	return ret;
}

char *
vpd_dir_get_default(const char * dir)
{
	return vpd_dir_get_subdirnum(dir, 0);
}

static void
vpd_dir_field_add_to_keys_file(const char * vpd_dir,
			       const vpd_key key)
{
	char * f;
	char * buf;
	size_t len;
	int err;
	char * out;
	char * t;

	if (NULL != (f = build_relative_filename(vpd_dir, VPD_LIST_FILE))) {
		if (!file_read_contents(f, &buf, &len, &err)) {
			buf = strdup("");
			len = 0;
		}

		out = NULL;

		if (0 == memcmp(key, "AX", sizeof(vpd_key)) &&
		    (NULL != (t = strstr(buf, "DS"))) &&
		    (('\0' == t[2]) || (' ' == t[2]))) {
			/* Adding AX, try to put after DS. */
			memcpy(t, "%s", 2);
			if (-1 == asprintf(&out, buf, "DS AX"))
				out = NULL;
		} else if (0 != memcmp(key, "YL", sizeof(vpd_key)) &&
			   (NULL != (t = strstr(buf, "YL"))) &&
			   (('\0' == t[2]) || (' ' == t[2]))) {
			/* Not adding YL, try to put before YL. */
			memcpy(t, "%s", 2);
			if (-1 != asprintf(&t, "%2.2s YL", key)) {
				if (-1 == asprintf(&out, buf, t))
					out = NULL;		
				free(t);
			}
		}

		if (NULL == out)
			if (-1 == asprintf(&out, "%s%2.2s ", buf, key))
				out = NULL;

		if (NULL != out) {
			file_write_contents(f, out, strlen(out), &err);
			free(out);
		}

		free(buf);
		free(f);
	}
}

extern bool
vpd_dir_field_ensure(const char * vpd_dir,
		     const vpd_key key,
		     const char * value)
{
	bool ret;
	char * f;
	struct stat buf;
	int err;

	ret = false;

	if (-1 != asprintf(&f, "%s/%s", vpd_dir, key)) {
		if (0 != lstat(f, &buf))
			if (file_write_contents(f,
						value, strlen(value),
						&err)) {
				vpd_dir_field_add_to_keys_file(vpd_dir, key);
				ret = true;
			}
		free(f);
	}

	return ret;
}
