/*
  pci_ethernet - Get PCI (and other) information for ethernet adapters.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2004
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: pci_ethernet.c,v 1.11 2005/02/22 22:33:23 martins Exp $
*/
  
#define _GNU_SOURCE

#include <stdio.h>
#include <unistd.h>

typedef __uint64_t u64;
typedef __uint32_t u32;
typedef __uint16_t u16;
typedef __uint8_t  u8;

#include <net/if.h>
#include <linux/sockios.h>
#include <linux/ethtool.h>
#include <net/if_arp.h>

#include <sys/ioctl.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"
#include "pci_ethernet.h"

void
pci_ethernet_fprint(FILE * f,
		    pci_ethernet_t * l)
{
	int i;

	while (NULL != l) {
		fprintf(f, "%.*s\n", IFNAMSIZ, l->name);
		fprintf(f, "\tbus\t%04lx:%02hhx:%02hhx:%02hhx\n",
			(unsigned long) l->dom, (unsigned char) l->bus,
			(unsigned char) l->dev, (unsigned char) l->fun);
		fprintf(f, "\tirq\t%hu\n", l->irq);
		fprintf(f, "\tstart\t%016llx\n",
			(unsigned long long) l->mem_start);
		fprintf(f, "\tend\t%016llx\n",
			(unsigned long long) l->mem_end);
		fprintf(f, "\tfw_ver\t%.32s\n", l->fw_version);
		
		fprintf(f, "\thwaddr\t");
		for (i=0; i < PCI_ETHERNET_HWADDR_LEN ; i++) {
			if (i > 0)
				fprintf(f, ":");
			fprintf(f, "%02hhx", l->hwaddr[i]);
		}
		fprintf(f, "\n");

		l = l->next;
	}
}

void
pci_ethernet_free(pci_ethernet_t * l)
{
	pci_ethernet_t * t;
	while (NULL != l) {
		t = l;
		l = l->next;
		free(t->name);
		free(t);
	}
}

static bool
pci_ethernet_line_scanner(const char * line,
			  void * x)
{
	char name[1024];
	char * s;
	int fd;
	struct ifreq ifr;
        struct ethtool_drvinfo drvinfo;
	unsigned int dom;
	unsigned char bus;
	unsigned char dev;
	unsigned char fun;
	pci_ethernet_t ** l;
	pci_ethernet_t * t;

	l = x;
	t = NULL;


	memset(name, 0, 1024);
	if (1 != sscanf(line, " %1023s", name))
		return true;

	if ((NULL == (s = memchr(name, ':', 1023))) || (s == name))
		return true;
	
	*s = '\0';
	
	if (-1 == (fd = socket(AF_INET, SOCK_DGRAM, 0)))
		return true;
	
	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, name, IFNAMSIZ);
	
	if ((-1 == ioctl(fd, SIOCGIFHWADDR, &ifr)) ||
	    (ARPHRD_ETHER != ifr.ifr_hwaddr.sa_family))
		goto done;
	
	t = calloc(1, sizeof(pci_ethernet_t));
	
	memcpy(t->hwaddr, ifr.ifr_hwaddr.sa_data, PCI_ETHERNET_HWADDR_LEN);
	
	memset(&drvinfo, 0, sizeof(struct ethtool_drvinfo));
	drvinfo.cmd = ETHTOOL_GDRVINFO;
	ifr.ifr_data = (caddr_t)&drvinfo;
	if (-1 == ioctl(fd, SIOCETHTOOL, &ifr))
		goto bail;
	
	if (4 == sscanf(drvinfo.bus_info,
			"%04x:%02hhx:%02hhx.%hhx",
			&dom, &bus, &dev, &fun))
		t->dom = dom;
	else if (3 == sscanf(drvinfo.bus_info,
			     "%02hhx:%02hhx.%hhx", 
			     &bus, &dev, &fun))
		t->dom = PCI_ETHERNET_NODOM;
	else
		goto bail;
	
	t->bus = bus;
	t->dev = dev;
	t->fun = fun;
	
	memcpy(t->fw_version, drvinfo.fw_version, 32);
	
	if (-1 == ioctl(fd, SIOCGIFMAP, &ifr))
		goto bail;
	
	t->irq       = (uint16_t) ifr.ifr_map.irq;
	t->mem_start = (uint64_t) ifr.ifr_map.mem_start;
	t->mem_end   = (uint64_t) ifr.ifr_map.mem_end;
	
	t->name = strndup(ifr.ifr_name, IFNAMSIZ);
	t->next = *l;
	*l = t;

 done:
	close(fd);
	return true;
 bail:
	free(t);
	goto done;
}

/* Extra level of functions to allow testing against fixed files. */
static pci_ethernet_t *
pci_ethernet_file(const char * filename)
{
	pci_ethernet_t * l = NULL;
	line_scanner_harness(filename, pci_ethernet_line_scanner, &l);
	return l;
}


pci_ethernet_t *
pci_ethernet()
{
	return pci_ethernet_file("/proc/net/dev");
}

#if 0
int
main(int argc, char ** argv)
{
	pci_ethernet_t * l = pci_ethernet();

	pci_ethernet_fprint(stderr, l);
	pci_ethernet_free(l);

        return 0;
}
#endif
