/*
  pci_ethernet_map - Print PCI (and other) information for ethernet adapters.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2004
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: pci_ethernet_map.c,v 1.3 2004/10/22 08:07:15 martins Exp $
*/

#include <stdbool.h>
#include <stdio.h>
#include <linux/pci_ids.h>
#include <net/if.h>
  
#include "pci_ethernet.h"
#include "proc_pci.h"

static void
print_item(const char * name,
	   uint32_t dom,  uint8_t bus, uint8_t dev, uint8_t fun,
	   const unsigned char * hwaddr)
{
	int i;

	printf("%04x:%02hhx:%02hhx.%hhx %.*s ",
	       (unsigned int)  dom,
	       (unsigned char) bus,
	       (unsigned char) dev,
	       (unsigned char) fun,
	       IFNAMSIZ, name);

	for (i=0; i < PCI_ETHERNET_HWADDR_LEN ; i++) {
		if (i > 0)
			printf(":");
		printf("%02hhx", hwaddr[i]);
	}
	printf("\n");
}

static bool
is_possible_match(pci_ethernet_t * e, proc_pci_list_t * p)
{
	memlist_t * m;

	if ((p->class == PCI_CLASS_NETWORK_ETHERNET) &&
	    (e->bus == (p->bus & 0xff))          &&
	    (e->dev == p->dev)                   &&
	    (e->fun == p->fun)                   &&
	    ((0 == e->irq) || (e->irq == (p->irq & 0xff)))) {

		/* Find first match on memory. */
		m = p->mem;
		while (NULL != m) {
			if ((0 == e->mem_start) ||
			    ((e->mem_start & 0xffffffff) ==
			     (m->start & 0xffffffff)))
				return true;
			m = m->next;
		}
	}

	return false;
}

static bool
pci_unique_ethernet_match(proc_pci_list_t * p, pci_ethernet_t * el)
{
	pci_ethernet_t * t;
	unsigned int num_matches = 0;

	for (t = el; NULL != t; t = t->next)
		if (is_possible_match(t, p)) {
			num_matches++;
			if (1 < num_matches)
				return false;
		}

	return (1 == num_matches);
}

static bool
ethernet_unique_pci_match(pci_ethernet_t * e, proc_pci_list_t * pl)
{
	proc_pci_list_t * t;
	unsigned int num_matches = 0;

	for (t = pl; NULL != t; t = t->next)
		if (is_possible_match(e, t)) {
			num_matches++;
			if (1 < num_matches)
				return false;
		}

	return (1 == num_matches);
}

void print_dodgy_matches(pci_ethernet_t * e, pci_ethernet_t * el)
{
	pci_ethernet_t * et;
	proc_pci_list_t * t;
	uint32_t dom;
	proc_pci_list_t * pl = proc_pci();

	for (t = pl; NULL != t; t = t->next)
		if (is_possible_match(e, t)) {
			dom = (uint32_t) (t->bus >> 8) & 0xffffff;
			if (pci_unique_ethernet_match(t, el))
				/* PCI address matches one NIC. */
				print_item(e->name,
					   dom, e->bus, e->dev, e->fun,
					   e->hwaddr);
			else
				/* Find a different NIC that matches
				   the address we're looking at.  If
				   this other NIC matches only that
				   address, then the address can't
				   belong to the current adapter -
				   otherwise print it.  */
				for (et = el; NULL != et; et = et->next)
					if ((e != et) &&
					    is_possible_match(et, t) &&
					    (!ethernet_unique_pci_match(et,
									pl)))
						print_item(e->name,
							   dom, e->bus,
							   e->dev, e->fun,
							   e->hwaddr);
		}

	proc_pci_list_free(pl);
}

int
main(int argc, char ** argv)
{
	pci_ethernet_t  * el;
	pci_ethernet_t  * t;

	if (NULL == (el = pci_ethernet()))
		return 1;

	for (t = el; NULL != t; t = t->next)
		if (PCI_ETHERNET_NODOM != t->dom)
			print_item(t->name,
				   t->dom, t->bus, t->dev, t->fun,
				   t->hwaddr);
		else
			print_dodgy_matches(t, el);

	pci_ethernet_free(el);

        return 0;
}
