/*
  pci_vpd_parse - Parse PCI VPD chunks.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: pci_vpd_parse.c,v 1.1 2005/01/17 07:05:58 martins Exp $
*/
  
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "growbuf.h"
#include "pci_vpd_parse.h"
#include "util.h"
#include "vpd.h"

static bool
is_printable(const unsigned char * s)
{
	while (*s != '\0') {
		if (!isprint((int) *s))
			return false;
		s++;
	}
	return true;
}

static bool
is_good(const vpd_list * kv)
{
	return (isprint(kv->key[0]) && isprint(kv->key[1]) &&
		(is_printable(kv->val) || (strncmp(kv->key, "DS", 2) == 0)) &&
		(kv->val[0] != '\0'));
}

static uint8_t
checksum_data(uint8_t init, uint8_t * data, size_t len)
{
	size_t i;
	for (i = 0; i < len ; i++) {
		init += data[i];
	}
	return init;
}

static uint8_t
parse_uchar(growbuf_iter * i)
{
	uint8_t ret;

	ret = growbuf_iter_current(i)[0];
	growbuf_iter_next(i);

	return ret;
}

static bool
parse_uchar_len(growbuf_iter * i, size_t * len)
{
	if (growbuf_iter_more(i)) {
		*len = (size_t) growbuf_iter_current(i)[0];
		growbuf_iter_next(i);
		return true;
	} else {
		return false;
	}
}

static bool
parse_uleshort_len(growbuf_iter * i, size_t * len)
{
	if (growbuf_iter_have(i, 2)) {
		*len = (size_t) (* growbuf_iter_current(i));
		growbuf_iter_next(i);
		*len += (size_t) (256 * (* growbuf_iter_current(i)));
		growbuf_iter_next(i);	
		return true;
	} else {
		return false;
	}
}

static bool
parse_uchar_length_data(growbuf_iter * i, size_t * len, char ** data)
{
	if (parse_uchar_len(i, len)) {
		if (growbuf_iter_have(i, *len)) {
			*data = growbuf_iter_current(i);
			growbuf_iter_move(i, *len);
			return true;
		} else {
			return false;
		}
	} else {
		return false;
	}
}

static bool
parse_uleshort_length_data(growbuf_iter * i, size_t * len, char ** data)
{
	if (parse_uleshort_len(i, len)) {
		if (growbuf_iter_have(i, *len)) {
			*data = growbuf_iter_current(i);
			growbuf_iter_move(i, *len);
		}
		return true;
	} else {
		return false;
	}
}

static bool
parse_label(growbuf_iter * i, vpd_key k)
{
	/* Can't use sizeof(k) here, since it gets coerced into a pointer. */
	if (growbuf_iter_have(i, sizeof(vpd_key))) {
		memcpy(k, growbuf_iter_current(i), sizeof(vpd_key));
		growbuf_iter_move(i, sizeof(vpd_key));
	}

	return true;
}	

static vpd_list *
parse_key_val(growbuf_iter * i)
{
	vpd_key k;
	char * val_data;
	size_t val_len;
	vpd_list * vl = NULL;

	if (parse_label(i, k) &&
	    parse_uchar_length_data(i, &val_len, &val_data))
		vl = vpd_list_new(k, val_data, val_len);

	return vl;
}
	
static void
parse_board_id(growbuf_iter * i, vpd_list ** vl, uint8_t * cs)
{
	char * name;
	size_t name_len;

	char * ds_key = "DS";
	char * invalid = "NOT_VALID_VPD";
	uint8_t t;
	char * yip;

	if (!parse_uleshort_length_data(i, &name_len, &name)) {
		/* This won't happen very often. */
		name = invalid;
		name_len = strlen(invalid);
	}

	t = (uint8_t) name_len;
	*cs = checksum_data(*cs, &t, sizeof(t));
	*cs = checksum_data(*cs, name, name_len);

	/* This works around a NUL character in name.  Yuck, huh? */
	if (NULL != (yip = memchr(name, '\0', name_len))) {
		name_len = (size_t) (yip - name);
	}

	*vl = vpd_list_append_key_value(*vl, ds_key, name, name_len);
}

static void
parse_resource_chunk(growbuf_iter * i, vpd_list ** vl, uint8_t * cs)
{
	size_t len;
	uint8_t * start;

	/* Some VPD chunk lengths are broken, but we want to try to
	 * trust the VPD, so we actually advance by the key-value
	 * pairs until we've exceeded the stated length...  but only
	 * if we're not going to walk off the end of the earth!
	 */
	if (parse_uleshort_len(i, &len)) {
		size_t len2;
		vpd_list * kv;
		*cs = checksum_data(*cs, (uint8_t *)&len, sizeof(len));

		start = growbuf_iter_current(i);

		while ((growbuf_iter_current(i) < (start + len)) &&
		       (NULL != (kv = parse_key_val(i))))

			if (is_good(kv))
				*vl = vpd_list_append(*vl, kv);
			else
				vpd_list_free(kv);
		
		len2 = growbuf_iter_current(i) - start;
		if (len != len2) {
			fprintf(stderr,
				"%s: stated length of chunk (0x%04zx) != "
				"actual length (0x%04zx).\n",
				vpd_program_name, len, len2);
		}
		
		*cs = checksum_data(*cs, start, len2);
	}
}

/* This only generates warnings. */
static void
do_checksum(growbuf_iter * i, uint8_t cs, size_t cs_len)
{
	switch (cs_len) {
	case 0:
		break;
	case 1:
		/* We can only handle a 1 byte checksum.  This is what
		 * is used in ibm,vpd files.  I've never seen another
		 * example.  */
		if (growbuf_iter_more(i)) {
			cs += parse_uchar(i);
			if (cs != 0) {
				fprintf(stderr,
					"%s: checksum=0x%02x "
					"incorrect for resource.\n",
					vpd_program_name, cs);
			}
		} else {
			fprintf(stderr,
				"%s: missing checksum data.\n",
				vpd_program_name);
		}
		break;
	default:
		fprintf(stderr,
			"%s: unknown checksum length=%zd.\n",
			vpd_program_name, cs_len);
	}
}

#define RTT_ID 0x02
#define RTT_RO 0x10
#define RTT_RW 0x11
#define RTT_END 0x0f

vpd_list *
pci_vpd_parse(growbuf * s)
{
	uint8_t tag, name;
	uint8_t cs = 0;
	
	growbuf_iter * i = growbuf_iter_new(s);
	vpd_list * ret = vpd_list_new_empty();
	
	do {
		/* What should happen here?  The VPD doesn't end
		 * correctly but we may have parsed some useful VPD.
		 * OK, let's print a message and break from the
		 * loop. */
		if (!growbuf_iter_more(i)) {
			fprintf(stderr, 
				"%s: pci_vpd_parse: incomplete VPD chunk.\n",
				vpd_program_name);
			break;
		}
		tag = parse_uchar(i);
		
		if ((tag & (1 << 7)) != 0) {
			/* Large Resource Data Type tag, length follows. */
			cs += tag;
			name = tag & ~(1 << 7);
			
			switch (name) {
			case RTT_ID:
				parse_board_id(i, &ret, &cs);
				break;
			case RTT_RO:
			case RTT_RW:
				parse_resource_chunk(i, &ret, &cs);
				break;
			default:
				fprintf(stderr, 
					"%s: internal error: "
					"Large Resource Name"
					"0x%02x not handled.\n",
					vpd_program_name, name);
				exit(1);
			}
		} else {
			name = (tag >> 3) & 0x0f;
			
			if (RTT_END == name) {
				do_checksum(i, cs, tag & 0x7);
			} else {
				fprintf(stderr, 
					"%s: internal error: "
					"Small Resource Name"
					"0x%02x not handled.\n",
					vpd_program_name, name);
				exit(1);
			}
		}
		
	} while (RTT_END != name);

	growbuf_iter_free(i);
	
	return ret;
}

