/* -*- mode: C -*-

  scsi_vpd_render - Render SCSI VPD components using templates.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: scsi_vpd_render.c,v 1.17 2005/02/22 04:14:46 martins Exp $
*/
  
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <regex.h>

#include "util.h"

#define TEMPLATE_FILE LSVPD_LIBDIR "/scsivpd.conf"

static void
usage()
{
	fprintf(stderr, "usage: %s indir outdir\n", vpd_program_name);
	exit(1);
}

static char *
hexify(const char * s, size_t len)
{
	int i;
	char * ret = malloc(len*2+1);
	for (i = 0; i < len ; i++) {
		snprintf(&ret[i*2], 3, "%.2X", (unsigned char) s[i]);
	}

	return ret;
}

static bool
dump_field_basic(const char * dir,
		 const char * key,
		 const char * data)
{
	int err;

	return file_relative_write_contents(dir, key,
					    data, strlen(data), &err);
}

static bool
dump_field(const char * dir,
	   const char * key,
	   const char * data,
	   size_t       len,
	   const char * dtype)
{
	bool ret;
	char * ldata;

	ret = false;

	/* Certain fields will require some munging. */
	/* Convert RL, Z0, Z7 to hex. */
	if (((strcmp(key, "RL") == 0) && (strcmp(dtype, "disk") == 0)) ||
	    (strcmp(key, "Z0") == 0) ||
	    (strcmp(key, "Z7") == 0))
		/* hexify returns a new NUL-terminated string. */
		ldata = hexify(data, len);
	else
		/* Make trimmed, NUL-terminated copy. */
		ldata = strndup_trim(data, len);

	
	if (NULL != ldata) {
		/* Write data to file. */
		ret = dump_field_basic(dir, key, ldata);
		free(ldata);
	}

	return ret;
}

#define ARRAY_LEN(x) (sizeof((x))/sizeof((x)[0]))

/* Very vaguely taken from linux-2.5.22/drivers/scsi/scsi.[ch]
 * There are only really 14 possible type, but filling it out to 16
 * removes the need for error checking. */
static const struct string_array scsi_device_types_short[] =
{
        {0x0, "disk"},
        {0x1, "tape"},
        {0x2, "printer"},
        {0x3, "processor"},
        {0x4, "worm"},
        {0x5, "cdrom"},
        {0x6, "scanner"},
        {0x7, "optical-disk"},
        {0x8, "jukebox"},
        {0x9, "communications"},
        {0xa, "graphics"},
        {0xb, "graphics"},
        {0xd, "enclosure"},
        {-1, NULL}
};


static const struct string_array scsi_device_types[] =
{
        {0x0, "Disk Drive"},
        {0x1, "Tape Drive"},
        {0x2, "Printer Device"},
        {0x3, "Processor Device"},
        {0x4, "Write-once Device"},
        {0x5, "CD-ROM Drive"},
        {0x6, "Scanner Device"},
        {0x7, "Optical Memory Device"},
        {0x8, "Medium Changer Device"},
        {0x9, "Communications Device"},
        {0xa, "Graphics Arts Pre-press Device"},
        {0xb, "Graphics Arts Pre-press Device"},
        {0xd, "Enclosure Services Device"},
        {-1, NULL}
};

#define SCSI_CDROM_DEFAULT "Other SCSI CD-ROM Drive"
static const struct string_array cdrom_device_types[] =
{
	{0, SCSI_CDROM_DEFAULT},
	{1, "OEM SCSI CD-ROM Drive"},
	{2, "SCSI CD-ROM Drive"},
	{3, "SCSI Multimedia CD-ROM Drive"},
	{4, "OEM SCSI DVD-ROM Drive"},
	{5, "SCSI DVD-ROM Drive"},
	{6, "SCSI DVD-RAM Drive"},
	{7, "OEM SCSI DVD-RAM Drive"},
	{-1, NULL}
};

static char *
make_ds_value(const char * data, unsigned char dtype, const char * dir)
{
	char * ret;
	char * data_0xc7;
	size_t len_0xc7;
	int err;
	char * s;
	const char * t;
	char * l;
	int type;

	ret = NULL;

	/* Get info for naming quirks, such as LVD, Multimedia. */
	file_relative_read_contents(dir, "0xc7", &data_0xc7, &len_0xc7, &err);

	/* Now do DS and type. */

	/* Check for SCSI *Multimedia* CD-ROM Drive. */
	if (dtype == 5) {
		type = ((NULL != data_0xc7) && (34 <= len_0xc7)) ?
			(int) data_0xc7[33] : 0;

		/* So it can be freed later. */
		s = strdup(string_array_lookup(cdrom_device_types,
					       SCSI_CDROM_DEFAULT,
					       type));
	} else {
		t = string_array_lookup(scsi_device_types,
					"Unknown Device",
					dtype);
		l = ((0 == dtype) && (NULL != data_0xc7) &&
		     (35 <= len_0xc7) && (7 == data_0xc7[34])) ?
			"LVD " : "";
		if (-1 == asprintf(&s, "%sSCSI %s", l , t))
			s = NULL;
	}

	if (NULL != s) {
		if (-1 == asprintf(&ret, "%s%s",
				   ((data[7] & 0x40) ? "32 Bit " :
				    ((data[7] & 0x20) ? "16 Bit " : "")),
				   s))
			ret = NULL;
		free(s);
	}

	if (NULL != data_0xc7)
		free(data_0xc7);
	
	return ret;
}

typedef struct {
	char * v_t_m;
	char * templates;
} get_templates_t;

/* Check if len bytes of template s are a regexp matching string vtm. */
static bool
template_matches(const char * s,
		 size_t len,
		 const char * vtm)
{
	bool ret;
	char * r;
	regex_t preg;

	ret = false;

	/* Got to copy since arg is const. */
	if (NULL != (r = malloc(len + 1))) {
		memcpy(r, s, len);
		r[len] = '\0';
		
		/* Is this the one? */
		if (0 == regcomp(&preg, r, REG_EXTENDED|REG_NOSUB)) {
			ret = (0 == regexec(&preg, vtm, 0, NULL, 0));
			regfree(&preg);
		}
		free(r);
	}

	return ret;
}

static bool
get_templates_line_scanner(const char * line,
			   void * data)
{
	bool ret;
	const char * t;
	int i;
	get_templates_t * inout;

	/* Default is we need to look at subsequent lines. */
	ret = true;
	inout = data;
	inout->templates = NULL;

	/* Check for comments. */
	if ((0 < strlen(line)) && ('#' != line[0])) {
		/* Find the 3rd semicolon. */
		i = 0;
		t = line;
		while ((i < 3) && (NULL != (t = strchr(t, ';')))) {
			i++;
			t++; /* Move past semi-colon. */
		}

		if ((3 == i) &&
		    (template_matches(line, t - 1 - line, inout->v_t_m))) {
			inout->templates = strdup(t);
			ret = false;
		}
	}

	return ret;
}

static char *
get_templates(const char * vendor,
	      const char * type,
	      const char * model)
{
	char * ret;
	get_templates_t inout;

	ret = NULL;

	if (-1 != asprintf(&inout.v_t_m, "%s;%s;%s", vendor, type, model)) {
		if (line_scanner_harness(TEMPLATE_FILE,
					 get_templates_line_scanner,
					 &inout))
			ret = inout.templates;

		free(inout.v_t_m);
	}

	return ret;
}

/* Like strchr, but replace char by '\0' and return pointer to next char. */
static char *
strchr_nul_plus_plus(char * s, char c)
{
	char * ret;

	ret = strchr(s, c);
	if (NULL != ret) {
		*ret = '\0';
		ret++;
	}

	return ret;
}

static void
process_template(char * template,
		 const char * indir,
		 const char * outdir,
		 const char * type)
{
	char * buf;
	size_t length;
	int err;
	int offset;
	char * s;
	char * k;
	char * v;
	int field_len;

	/* NUL-terminate page name. */
	s = template;
	s = strchr_nul_plus_plus(s, '=');

	/* Can't read file?  No problem: template might be too specific. */
	if ((NULL != s) &&
	    (file_relative_read_contents(indir, template,
					 &buf, &length, &err))) {

		/*
		 * Templates skip 1st 32 bytes of standard inquiry,
		 * since this is standard and handled elsewhere.
		 */
		offset = (strcmp(template, "inquiry") == 0) ? 32 : 0;

		while ((NULL != s) && (offset < length)) {
			/* Get field key. */
			k = s;
			s = strchr_nul_plus_plus(s, ':');
			if (NULL != s) {
				/* Get field value length. */
				v = s;
				s = strchr_nul_plus_plus(s, ',');
				field_len = atoi(v);

				/*
				 * Dump value to file if desired and
				 * have enough bytes in buf.
				 */
				if ((k[0] != '_') &&
				    (offset + field_len <= length)){
					dump_field(outdir, k, buf + offset,
						   field_len, type);
				}

				/* Done, move along. */
				offset += field_len;
			}
		}
		free(buf);
	}
}

static void
process_templates(char * templates,
		  const char * indir,
		  const char * outdir,
		  const char * type)
{
	char * s;
	char * t;

	s = templates;

	while (NULL != s) {
		t = s;

		s = strchr_nul_plus_plus(s, ';');
		process_template(t, indir, outdir, type);
	}
}

int
main(int argc, char **argv)
{
	char * mf;
	char * tm;
	char * data;
	size_t len;
	int err;
	unsigned char dtype;
	const char * dtype_short;
	char * dtype_str;
	char * templates;
	
	vpd_program_name = argv[0];

	if (argc != 3) {
		usage();
	}

	/* Grab contents of inquiry file. */
	file_relative_read_contents(argv[1], "inquiry", &data, &len, &err);

	/* Validate data. */
	if (len < 32) {
		fprintf(stderr, "%s: short standard inquiry data in %s/%s\n",
			vpd_program_name, argv[1], "inquiry");
		exit(1);
	}

	dtype = data[0] & 0x1F;
	dtype_short = string_array_lookup(scsi_device_types_short,
					  "unknown",
					  dtype);

	/* Parse out standard fields. */
	dump_field(argv[2], "Z0", data, 8, dtype_short);

	mf = strndup_trim(data + 8, 8);
	dump_field_basic(argv[2], "MF", mf);

	tm = strndup_trim(data + 16, 16);
	dump_field_basic(argv[2], "TM", tm);

	/* Now do DS and type. */
	dtype_str = make_ds_value(data, dtype, argv[1]);
	dump_field(argv[2], "DS", dtype_str, strlen(dtype_str), dtype_short);
	free(dtype_str);

	free(data);

	/* "type" is used externally. */
	if (file_relative_write_contents(argv[1], "type",
					 dtype_short, strlen(dtype_short),
					 &err)) {

		/* Parse out extra fields using templates. */
		templates = get_templates(mf, dtype_short, tm);
		if (templates) {
			process_templates(templates, argv[1], argv[2],
					  dtype_short);
			free(templates);
		}
	}

	free(mf);
	free(tm);

	return 0;
}
